"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
// Built-in
const http = require("http");
const path = require("path");
const fs = require("fs");
// Light-weight
const uuid = require("uuid/v4");
// Used immediately
const express = require("express");
const compression = require("compression");
const socketio = require("socket.io");
const middleware = require("./middleware");
// Local
const project_1 = require("./project");
const jsonLines_1 = require("./jsonLines");
const liveSharing_1 = require("./liveSharing");
var OutdatedPackage;
(function (OutdatedPackage) {
    function fromTableRow(row) {
        return {
            name: row[0],
            current: row[1],
            wanted: row[2],
            latest: row[3],
            type: row[4],
        };
    }
    OutdatedPackage.fromTableRow = fromTableRow;
})(OutdatedPackage = exports.OutdatedPackage || (exports.OutdatedPackage = {}));
var ServerCommand;
(function (ServerCommand) {
    ServerCommand["addPackage"] = "addPackage";
    ServerCommand["removePackage"] = "removePackage";
    ServerCommand["publishPackage"] = "publishPackage";
    ServerCommand["reloadClients"] = "reloadClients";
    ServerCommand["updatePackageInfo"] = "updatePackageInfo";
    ServerCommand["setIsDocumentDirty"] = "setIsDocumentDirty";
    ServerCommand["upgradeDependency"] = "upgradeDependency";
    ServerCommand["upgradeFramerLibrary"] = "upgradeFramerLibrary";
    ServerCommand["updateCanvasData"] = "updateCanvasData";
    ServerCommand["updateCanvasModule"] = "updateCanvasModule";
    ServerCommand["startLiveSharing"] = "startLiveSharing";
    ServerCommand["stopLiveSharing"] = "stopLiveSharing";
    ServerCommand["joinRoom"] = "joinRoom";
    ServerCommand["updatePreview"] = "updatePreview";
    ServerCommand["metrics"] = "metrics";
    ServerCommand["buildStandaloneExport"] = "buildStandaloneExport";
    ServerCommand["setProjectDir"] = "setProjectDir";
})(ServerCommand = exports.ServerCommand || (exports.ServerCommand = {}));
function setLongCache(res) {
    res.setHeader("Cache-Control", "public, max-age=36000");
    res.setHeader("Pragma", "");
    res.setHeader("Surrogate-Control", "");
}
class Server {
    constructor(config) {
        this.config = config;
        this._state = {
            url: undefined,
            filename: undefined,
            running: false,
            packages: [],
            outdated: [],
            packageErrors: [],
            serverErrors: [],
            tasks: {},
            isDocumentDirty: false,
            liveSharing: liveSharing_1.LiveSharing.initialState(),
            packagesInfo: {},
            hasProject: false,
        };
        this.hostImages = true;
        this.previewData = null;
        this.rawCanvasData = null;
        const app = express();
        this.expressApp = app;
        this.http = http.createServer(app);
        this.io = socketio(this.http);
        this.liveSharing = new liveSharing_1.LiveSharing("http://framer.live");
        this.liveSharing.on("statusUpdate", () => {
            this.setState({ liveSharing: this.liveSharing.state });
        });
        this.liveSharing.on("error", error => {
            Server.logError(`Live Preview Error: ${error}`);
        });
        app.use(middleware.nocache);
        app.use(middleware.logging);
        app.use(middleware.cors);
        app.use(compression());
        const socketPath = path.join(__dirname, "..", "node_modules/socket.io-client/dist");
        app.use("/_socket", express.static(socketPath));
        // Use for debugging, example usage
        app.get("/_loader", (req, res) => {
            res.send(`

            <script src="/_socket/socket.io.js"></script>
            <script>var socket = io(); socket.on("script", function(script) {

                const require = (name) => {
                    throw Error("Component loader: Can't require '" + name + "' likely an external in webpack")
                }

                const fn = new Function("exports", "require", script)
                const exportedModules = {}

                fn(exportedModules, require)

                console.log("package",  exportedModules)

            });
            </script>
          `);
        });
        app.use("/_proxy", function (req, res) {
            req.pipe(lazyLoadRequest()(req.url.slice(1))).pipe(res);
        });
        Server.instance = this;
        this.setup();
    }
    get state() {
        if (!this.config)
            return this._state;
        // If a proxy is forwarding requests to this server, the public URLs will be different.
        const { proxyDocumentURL, proxyImageBaseURL, proxyURL } = this.config;
        const state = Object.assign({}, this._state);
        // TODO: Remove check for existing documentURL (see updateURLs method).
        if (proxyDocumentURL && this._state.documentURL) {
            state.documentURL = proxyDocumentURL;
        }
        if (proxyImageBaseURL)
            state.imageBaseURL = proxyImageBaseURL;
        if (proxyURL)
            state.url = proxyURL;
        return state;
    }
    static logError(error) {
        const instance = Server.instance;
        const queue = Server.errorsQueue;
        if (!instance) {
            queue.push(error);
            return;
        }
        // Sending the queue first
        if (queue.length) {
            queue.forEach((queuedError) => {
                instance.sendError(queuedError);
            });
            queue.length = 0;
        }
        instance.sendError(error);
    }
    setProject(project) {
        this.project = project;
        this.commandHandler = project.commandHandler;
        const app = this.expressApp;
        // must make sure design images are fully cachable
        app.use("/design/images", express.static(path.join(project.path, "design", "images"), {
            setHeaders: setLongCache,
        }));
        app.use("/.cache", express.static(path.join(project.path, ".cache"), {
            setHeaders: setLongCache,
        }));
        // normal images are under user control, should invalidate after a while
        app.use("/images", express.static(path.join(project.path, "images"), {
            setHeaders: function (res) {
                res.setHeader("Cache-Control", "public, must-revalidate, max-age=36000");
                res.setHeader("Pragma", "");
                res.setHeader("Surrogate-Control", "");
            },
        }));
        // redirect / and /index.html to preview/preview.html
        if (project.staticPath) {
            // This works because the static middleware will call next when not found
            app.get("/", function (req, res) {
                if (req.path === "/" || req.path === "/index.html") {
                    res.redirect("/preview/");
                }
            });
            // Needed to make the route above work
            app.use("/preview", express.static(path.join(project.staticPath, "Vekter"), { index: "preview.html" }));
            app.use("/_app/resources", express.static(project.staticPath));
        }
        app.use("/", express.static(project.path));
        const lazyMulter = lazyLoadMulter();
        const metadataUploadHandler = lazyMulter({ storage: lazyMulter.memoryStorage() }).single("resource");
        app.post("/upload/metadata/:resourceFilename", metadataUploadHandler, (req, res) => {
            const resourceFilename = path.basename(req.params.resourceFilename).toLowerCase();
            const isReadme = "readme.md" === resourceFilename;
            const dest = isReadme
                ? path.join(project.path, "README.md")
                : path.join(project.path, project.config.paths.metadata, resourceFilename);
            fs.writeFile(dest, req.file.buffer, err => {
                if (err) {
                    // tslint:disable-next-line:no-console
                    console.log(err);
                    res.sendStatus(400);
                    return;
                }
                res.sendStatus(200);
            });
        });
        this.updateURLs();
    }
    setup() {
        this.io.on("connect", socket => {
            Object.keys(ServerCommand).forEach((command) => {
                socket.on(command, info => {
                    if (!this.handleServerCommand(socket, command, info)) {
                        this.commandHandler && this.commandHandler(command, info);
                    }
                });
            });
            socket.on("resetServerErrors", () => this.setState({ serverErrors: [] }));
            // Send the state over the just connected socket
            socket.emit("state", this.state);
            // Send the script at load (for dev), but only to the socket connecting!
            this.project && this.project.sendScript(socket);
            if (this.state.serverErrors.length) {
                socket.emit("serverErrors", { errors: this.state.serverErrors });
            }
        });
    }
    // Returns true if handled by server
    handleServerCommand(socket, command, info) {
        switch (command) {
            // Hijack commands that don’t really involve the project
            case "startLiveSharing": {
                this.liveSharing.start(this.httpAddress().port);
                return true;
            }
            case "stopLiveSharing": {
                this.liveSharing.stop();
                return true;
            }
            case "joinRoom": {
                if (!info.room)
                    return true;
                socket.join(info.room);
                if (info.room === "preview") {
                    if (this.rawCanvasData) {
                        socket.emit("updateCanvasData", this.rawCanvasData);
                    }
                    if (this.previewData) {
                        socket.emit("updatePreview", this.previewData);
                    }
                }
                return true;
            }
            case "updatePreview": {
                this.previewData = info.previewData || null;
                this.io.to("preview").emit("updatePreview", info.previewData);
                return true;
            }
            case "updateCanvasData": {
                this.rawCanvasData = info.canvasData || null;
                this.io.to("preview").emit("updateCanvasData", info.canvasData);
                return false; // Because we still want the project to handle this command too!
            }
            case "metrics": {
                this.io.emit("metrics", { type: info.type });
                return true;
            }
            case "buildStandaloneExport": {
                if (!this.project || !info.path)
                    return true;
                this.project
                    .buildStandalone(info.path, info.previewData)
                    .then(() => {
                    socket.emit("buildStandaloneExportDone", "ok");
                })
                    .catch(error => {
                    // tslint:disable-next-line:no-console
                    console.log(error);
                    socket.emit("buildStandaloneExportDone", error && error.message ? error.message : error);
                });
                return true;
            }
            case "setProjectDir": {
                const { dir, hostImages } = info;
                if (!dir)
                    return true;
                if (!this.config)
                    return true;
                this.config.projectPath = dir;
                if (hostImages !== undefined) {
                    this.hostImages = hostImages;
                }
                const project = new project_1.Project(this.config);
                project._server = this;
                this.setProject(project);
                project.onServerStart();
                (() => __awaiter(this, void 0, void 0, function* () {
                    yield project.install();
                    project.watch();
                    yield project.updatePackageInfo();
                }))();
                return true;
            }
            default: {
                return false;
            }
        }
    }
    start(port = 3000) {
        let tryPort = typeof port === "number" ? port : port[0];
        const maxPort = typeof port === "number" ? port : port[1];
        // When setting a host (which we want, to prevent leaking info to the network),
        // we also need to set “exclusive”, else the server will end up pooling with other
        // servers.
        const listenOptions = { host: "127.0.0.1", exclusive: true };
        return new Promise((resolve, reject) => {
            const onError = error => {
                if (error.code === "EADDRINUSE" && tryPort < maxPort) {
                    this.http.listen(Object.assign({}, listenOptions, { port: ++tryPort })); // Retry
                }
                else {
                    this.http.removeListener("error", onError);
                    this.http.removeListener("listening", onListening);
                    reject(error);
                }
            };
            const onListening = () => {
                const { address, port: httpPort } = this.httpAddress();
                this.http.removeListener("error", onError);
                this.setState({
                    url: `http://${address}:${httpPort}`,
                    filename: this.project && path.basename(this.project.path),
                    running: true,
                });
                this.liveSharing.enable();
                this.updateURLs();
                resolve(httpPort);
            };
            this.http.on("error", onError);
            this.http.once("listening", onListening);
            this.http.listen(Object.assign({}, listenOptions, { port: tryPort }));
        });
    }
    hasPackageInfo(packageId) {
        return packageId in this._state.packagesInfo;
    }
    addPackageInfo(info) {
        const packageIds = Object.keys(info);
        const newPackages = packageIds.filter(packageId => !this.hasPackageInfo(packageId));
        if (newPackages.length === 0)
            return;
        const packagesInfo = Object.assign({}, this._state.packagesInfo);
        newPackages.forEach(newPackgeId => {
            let value = info[newPackgeId];
            if (value !== null && typeof value !== "object") {
                value = null;
            }
            packagesInfo[newPackgeId] = value;
        });
        this.setState({ packagesInfo });
    }
    httpAddress() {
        // We use the APIs in a way that this will never be a string
        return this.http.address();
    }
    registerTask(name, task, context) {
        return __awaiter(this, void 0, void 0, function* () {
            const taskId = uuid();
            this.addTaskToState(taskId, name, context);
            try {
                yield task;
            }
            finally {
                this.removeTaskFromState(taskId);
            }
        });
    }
    runAsTask(name, task, context) {
        return __awaiter(this, void 0, void 0, function* () {
            // Although registerTask is async, we intentionally don't wait for it here
            this.registerTask(name, task, context);
            return task;
        });
    }
    addTaskToState(id, name, context) {
        const tasks = Object.assign({}, this.state.tasks);
        tasks[id] = context ? { name, context } : { name };
        this.setState({ tasks });
    }
    removeTaskFromState(id) {
        const tasks = Object.assign({}, this.state.tasks);
        delete tasks[id];
        this.setState({ tasks });
    }
    setState(state) {
        Object.assign(this._state, state);
        // Log to stdout, used by the app (at least for the first state)
        // tslint:disable-next-line:no-console
        console.log(JSON.stringify(this._state));
        // Send the state over all connected sockets
        this.io.emit("state", this.state);
    }
    sendError(error) {
        try {
            const yarnErrors = jsonLines_1.JSONLines.parse(error);
            let errors = yarnErrors
                // Keep only error type
                .filter(item => item.type === "error")
                // Filter out errors summary 'Found N errors.'
                .filter(item => !item.data.includes("Found "));
            if (!yarnErrors.length && typeof error === "string" && error.length) {
                // Filter out a Yarn bug we often run into
                if (!error.includes("/node_modules/yarn/lib/cli.js")) {
                    errors = [{ data: error, type: "error" }];
                }
            }
            this.setState({ serverErrors: this.state.serverErrors.concat(errors) });
            this.io.emit("serverErrors", { errors });
        }
        catch (err) {
            // tslint:disable-next-line:no-console
            console.log("Error happened while sending error!");
        }
    }
    updateURLs() {
        if (!this.project || !this.project.path || !this.state.url)
            return;
        const projectPath = this.project.path;
        const { url } = this.state;
        let documentURL;
        if (fs.existsSync(path.join(projectPath, "design/document.json"))) {
            // TODO: Add another state property that says if document.json
            //       file exists, and always send this URL.
            documentURL = `${url}/design/document.json`;
        }
        const imageBaseURL = `${this.hostImages ? url : pathToFileURL(projectPath)}/design/images/`;
        this.setState({ documentURL, imageBaseURL, hasProject: true });
    }
}
Server.errorsQueue = [];
exports.Server = Server;
function pathToFileURL(str) {
    if (typeof str !== "string") {
        throw new Error("Expected a string");
    }
    let pathName = path.resolve(str).replace(/\\/g, "/");
    // Windows drive letter must be prefixed with a slash
    if (pathName[0] !== "/") {
        pathName = "/" + pathName;
    }
    return encodeURI("file://" + pathName);
}
// Lazy loading for start-up performance
const lazyLoadModule = (name) => () => require(name);
const lazyLoadRequest = lazyLoadModule("request");
const lazyLoadMulter = lazyLoadModule("multer");
